import asyncio
from contextlib import contextmanager
from typing import Any, Generator, Iterator


class PubSub:
    _instance = None

    def __init__(self):
        self.queues: list[asyncio.Queue] = []

    @classmethod
    def instance(cls) -> 'PubSub':
        if cls._instance == None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    @contextmanager
    def subscribe(cls) -> Iterator[asyncio.Queue]:
        queue = asyncio.Queue()

        instance = cls.instance()
        instance.queues.append(queue)

        try:
            yield queue
        finally:
            instance.queues.remove(queue)
            del queue

    @classmethod
    async def publish(cls, msg: Any):
        for queue in cls.instance().queues:
            await queue.put(msg)



from importlib import resources
from pathlib import Path
from typing import Callable, Awaitable, Optional

from aiohttp import web

from ttun.pubsub import PubSub
from ttun.types import Config, RequestData

BASE_DIR = Path(__file__).resolve().parent


def no_print(*args, **kwargs):
    pass


class Server:
    def __init__(self, config: Config, on_resend: Callable[[RequestData], Awaitable]):
        self.config = {
            **config,
            'assets': '/assets/'
        }

        self.on_resend = on_resend

        self.app = web.Application()
        with resources.path(__package__, 'staticfiles') as staticfiles:
            self.app.add_routes([
                web.get('/inspect/', self.inspect_socket),
                web.get('/config/', self.get_config),
                web.post('/resend/', self.resend),
                web.get('/', self.root),
                web.static('/', staticfiles),
            ])

    async def run(self):
        port = 4040
        while True:
            try:
                await web._run_app(self.app, port=port)
            except OSError:
                port += 1

    async def root(self, request: web.Request):
        with resources.path(__package__, 'staticfiles') as staticfiles:
            return web.Response(
                body=resources.read_text(f'{__package__}.staticfiles', f'index.html'),
                content_type='text/html'
            )

    async def get_config(self, request: web.Request):
        return web.json_response(
            self.config,
            headers={
                'Access-Control-Allow-Origin': '*'
            }
        )

    async def resend(self, request: web.Request):
        await self.on_resend(await request.json())
        return web.json_response()

    async def inspect_socket(self, request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        with PubSub.subscribe() as subscription:
            while message := await subscription.get():
                await ws.send_json(message, False)

        return ws



import asyncio
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from asyncio import FIRST_EXCEPTION

from ttun.client import Client
from asyncio.exceptions import TimeoutError, CancelledError

from ttun.inspect_server import Server
from ttun.settings import SERVER_HOSTNAME, SERVER_USING_SSL

inspect_queue = asyncio.Queue()

def main():
    parser = ArgumentParser(prog='ttun', formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'port',
        help='The local port to expose'
    )
    parser.add_argument(
        '--server',
        default=f'{"wss" if SERVER_USING_SSL else "ws"}://{SERVER_HOSTNAME}',
        help='The hostname of the ttun server',
    )
    parser.add_argument(
        '-s', '--subdomain',
        default=None,
        help = 'The subdomain of the ttun tunnel',
    )
    args = parser.parse_args()

    client = Client(
        port=args.port,
        subdomain=args.subdomain,
        server=args.server,
    )

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    loop.run_until_complete(client.connect())

    server = Server(config=client.config, on_resend=client.proxyRequest)
    tasks = {
        loop.create_task(client.handle_messages()),
        loop.create_task(server.run())
    }

    print(client.config['url'])
    try:
        loop.run_until_complete(asyncio.wait(tasks, return_when=FIRST_EXCEPTION))
    except (CancelledError, TimeoutError):
        for task in tasks:
            task.cancel()
        loop.close()
    except KeyboardInterrupt:
        pass

if __name__ == '__main__':
    main()
